using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Hashes;
using Hashes.StringHashes;

namespace ECollections
{
    /// <summary>
    /// A fast, memory efficient Dictionary.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class Dictionary<TKey, TValue> where TValue : IEquatable<TValue>
    {
        #region Constructors
        /// <summary>
        /// Capacity specifies the initial size of the Dictionary. The Dictionary grows too accomodate more entries as they are added.
        /// HashProviderA and HashProviderB are distinct hash providers for type TKey.
        /// </summary>
        /// <param name="capacity"></param>
        /// <param name="hashProviderA"></param>
        /// <param name="hashProviderB"></param>
        public Dictionary(int capacity, IHashProvider<TKey> hashProviderA, IHashProvider<TKey> hashProviderB)
        {
            Constructor(capacity, hashProviderA, hashProviderB, null);
        }
        public Dictionary(int capacity, IHashProvider<TKey> hashProviderA, IHashProvider<TKey> hashProviderB, IDictionary<TKey, TValue> dictionary)
        {
            Constructor(capacity, hashProviderA, hashProviderB, dictionary);
        }

        private void Constructor(int capacity, IHashProvider<TKey> hashProviderA, IHashProvider<TKey> hashProviderB,IDictionary<TKey, TValue> dictionary)
        {
            // Minimum capacity of Dictionary = 1024 entries
            _capacity = (capacity > _minCapacity ? GetNextCapacity(capacity) : 1024);
            _capacityLess1 = _capacity - 1;
            _entries = new DictionaryEntry<TKey,TValue>[_capacity];
            // Initialize the space map
            _spaceMap = new BitArray(_capacity);

            // If no hash provider is given, use default hash provider
            if (hashProviderA == null || hashProviderB == null)
                throw new ArgumentNullException((hashProviderA == null && hashProviderB == null) ? ("hashProviderA, hashProviderB") : ((hashProviderA == null) ? "hashProviderA" : "hashProviderB"));

            _hashProviderA = hashProviderA;
            _hashProviderB = hashProviderB;
            
            // If a dictionary is provided, add its members to present Dictionary
            if (dictionary != null)
                AddCollection(dictionary);
        }
        #endregion

        #region Private Members
        // Used for thread synchronization
        private object _syncRoot = new object();
        // Number of items present in Dictionary
        private int _count;
        // Hash Providers
        private IHashProvider<TKey> _hashProviderA;
        private IHashProvider<TKey> _hashProviderB;
        // Dictionary entries
        private DictionaryEntry<TKey, TValue>[] _entries;
        // The number of hashes to use by default. A higher number will provide higher load factor. A lower number will provide greater performance.
        private int _numHashes = 307;
        // Mininum Capacity of Dictionary
        private readonly int _minCapacity = 1024;
        // The space map for the dictionary
        private BitArray _spaceMap;
        // Capacity
        private int _capacity;
        private int _capacityLess1;
        #endregion

        #region Public Methods
        /// <summary>
        /// This methods adds a given Key value pair to the Dictionary. Keys need to be unique. Values need not be unique.
        /// Complexity Average Case: O(log((n + 1/)n)), Worst Case: O(n)
        /// </summary>
        /// <param name="key"></param>
        /// <param name="keyValue"></param>
        public void Add(TKey key, TValue keyValue)
        {
            if (key == null)
                throw new ArgumentNullException("key");

            if (this.ContainsKey(key))
                throw new ArgumentException("Key already exists, key: " + key.ToString(), "key");

            AddInternal(key, keyValue);
        }
        /// <summary>
        /// Removes a given Key-Value pair from the dictionary. If the specified key does not exist in the Dictionary, False is returned.
        /// Complexity O(1)
        /// </summary>
        /// <param name="key"></param>
        public bool Remove(TKey key)
        {
            if (key == null)
                throw new ArgumentNullException("key");

            int index = ContainsKeyInternal(key);
            // If key is present in dictionary, a value >= 0 is returned
            if (index >= 0)
            {
                lock (_syncRoot)
                {
                    _spaceMap.Set(index, false);
                    _count--;
                }
                return true;
            }
            else
                return false;
            // throw new Exception("Key does not exist in dictionary");   
        }

        /// <summary>
        /// Determines whether a given key exists in the Dictionary. 
        /// Complexity O(1)
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public bool ContainsKey(TKey key)
        {
            if (key == null)
                throw new ArgumentNullException("key");

            if (ContainsKeyInternal(key) >= 0)
                return true;
            else
                return false;
        }
        /// <summary>
        /// Determines whether a given Value exists in the Dictionary
        /// Complexity O(n)
        /// </summary>
        /// <param name="val"></param>
        /// <returns></returns>
        public bool ContainsValue(TValue val)
        {
            for (int i = 0; i < _entries.Length; i++)
            {
                if (_spaceMap[i] && _entries[i].Value.Equals(val))
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Clears all the Key-Value pairs from the Dictionary.
        /// Complexity O(1)
        /// </summary>
        public void Clear()
        {
            _entries = new DictionaryEntry<TKey, TValue>[_minCapacity];
            _spaceMap.SetAll(false);
        }
        /// <summary>
        /// Gets the number of items present in the Dictionary
        /// </summary>
        public int Count
        {
            get { return _count; }
        }
        /// <summary>
        /// Gets the maximum number of items the Dictionary can contain without resizing
        /// </summary>
        public int Capacity
        {
            get { return _capacity; }
        }
        /// <summary>
        /// Gets the first HashProvider being used by the Dictionary
        /// </summary>
        public IHashProvider<TKey> HashProviderA
        {
            get { return _hashProviderA; }
        }

        /// <summary>
        /// Gets the second HashProvider being used by the Dictionary
        /// </summary>
        public IHashProvider<TKey> HashProviderB
        {
            get { return _hashProviderB; }
        }   

        /// <summary>
        /// The number of hashes to generate for insertions / lookups in the Dictionary.
        /// The default value is 31. Possible values range from 5 to 277 inclusive. A lower value leads to greater performance at the cost of a low load factor.
        /// A higher value leads to lesser performance but higher load factor.
        /// </summary>
        public int NumHashes
        {
            get { return _numHashes; }
            set { _numHashes = ((value < 16) && (value > 278) ? value : 31); }
        }
        // Indexer
        public TValue this[TKey key]
        {
            get
            {
                if (key == null)
                    throw new ArgumentNullException("key");

                int index;
                if ((index = ContainsKeyInternal(key)) >= 0)
                    return _entries[index].Value;
                else
                    return default(TValue);
            }
            set
            {
                int index;
                if ((index = ContainsKeyInternal(key)) >= 0)
                    _entries[index].Value = value;
                else
                    throw new KeyNotFoundException("Key not found. Key: " + key);
            }
        }
        #endregion

        #region Utility Functions
        /// <summary>
        /// The internal method used by Add(string, T) method.
        /// </summary>
        /// <param name="key"></param>
        /// <param name="val"></param>
        private void AddInternal(TKey key, TValue val)
        {
            uint hashA, hashB;
            uint hash;
            bool inserted = false;
            hashA = HashProviderA.GetHashCode(key);
            hashB = HashProviderB.GetHashCode(key);
            
            lock (_syncRoot)
            {
                for (int i = 0; i < NumHashes; i++)
                {
                    hash = hashA + _primes[i] * hashB;
                    if (!_spaceMap[(int)(hash & _capacityLess1)])
                    {
                        _entries[hash & (this.Capacity - 1)] = new DictionaryEntry<TKey, TValue>(key, val);
                        inserted = true;
                        _spaceMap[(int)(hash & _capacityLess1)] = true;
                        break;
                    }
                }
                if (!inserted)
                {
                    while (!inserted)
                    {
                        Resize();
                        for (int i = 0; i < NumHashes; i++)
                        {
                            hash = hashA + _primes[i] * hashB;
                            if (!_spaceMap[(int)(hash & _capacityLess1)])
                            {
                                _entries[hash & _capacityLess1] = new DictionaryEntry<TKey, TValue>(key, val);
                                inserted = true;
                                _spaceMap[(int)(hash & _capacityLess1)] = true;
                                break;
                            }
                        }
                    }
                }
                _count++;
            }
        }

        /// <summary>
        /// The internal method used by ContainsKey(string) method.
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        private int ContainsKeyInternal(TKey key)
        {
            uint hashA, hashB;
            uint hash;
            hashA = HashProviderA.GetHashCode(key);
            hashB = HashProviderB.GetHashCode(key);
            for (int i = 0; i < NumHashes; i++)
            {
                hash = hashA + _primes[i] * hashB;
                if (!_spaceMap[(int)(hash & _capacityLess1)])
                    break;
                else if (_entries[hash & _capacityLess1].Key.Equals(key))
                    return (int)(hash & _capacityLess1);
            }
            return -1;
        }

        /// <summary>
        /// The internal method called to resize the _entries array.
        /// Complexity O(n)
        /// </summary>
        private void Resize()
        {
            _capacity = _capacity << 1;
            _capacityLess1 = _capacity - 1;
            DictionaryEntry<TKey, TValue>[] newEntries = new DictionaryEntry<TKey, TValue>[_capacity];
            BitArray newSpaceMap = new BitArray(_capacity);
            uint hash, hashA, hashB;

            for (int i = 0; i < _entries.Length; i++)
            {
                if (_spaceMap[i])
                {
                    hashA = HashProviderA.GetHashCode(_entries[i].Key);
                    hashB = HashProviderB.GetHashCode(_entries[i].Key);

                    for (int j = 0; j < NumHashes; j++)
                    {
                        hash = hashA + _primes[j] * hashB;
                        if (! newSpaceMap[(int)(hash & _capacityLess1)])
                        {
                            newEntries[hash & _capacityLess1] = _entries[i];
                            newSpaceMap[(int)(hash & _capacityLess1)] = true;
                            break;
                        }
                    }
                }
            }

            _entries = newEntries;
            _spaceMap = newSpaceMap;
        }

        /// <summary>
        /// Returns the next Prime greater than the given number.
        /// </summary>
        /// <param name="current"></param>
        /// <returns></returns>
        private int GetNextPrime(int current)
        {
            bool flag;
            current++;
            while (true)
            {
                flag = true;
                for (int i = 0; i < _primes.Length; i++)
                {
                    if (current % _primes[i] == 0)
                    {
                        flag = false;
                        break;
                    }
                }
                if (flag)
                {
                    for (int i = 2031; i < Math.Sqrt(current) + 1; i += 2)
                    {
                        if (current % i == 0)
                        {
                            flag = false;
                            break;
                        }
                    }
                }
                if (flag)
                    break;
                else
                    current++;
            }
            return current;
        }

        /// <summary>
        /// Returns the next capacity greater than the given number. Capacities are powers of 2.
        /// </summary>
        /// <param name="currentCapacity"></param>
        /// <returns></returns>
        private int GetNextCapacity(int currentCapacity)
        {
            for (int i = 10; i < 31; i++)
            {
                if ((2 << i) >= currentCapacity)
                    return 2 << i;
            }
            throw new Exception("Capacity could not be found for currentCapacity: " + Capacity);
        }

        /// <summary>
        /// Add given Dictionary<string,T> to present collection
        /// </summary>
        /// <param name="collection"></param>
        private void AddCollection(IDictionary<TKey, TValue> collection)
        {
            foreach (TKey key in collection.Keys)
            {
                this.Add(key, collection[key]);
            }
            _count += collection.Count;
        }

        /// <summary>
        /// Some primes to help speed up prime calculation
        /// </summary>
        private readonly uint[] _primes = new uint[]   {
                                                2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541, 547, 557, 563, 569, 571, 577, 587, 593, 599, 601, 607, 613, 617, 619, 631, 641, 643, 647, 653, 659, 661, 673, 677, 683, 691, 701, 709, 719, 727, 733, 739, 743, 751, 757, 761, 769, 773, 787, 797, 809, 811, 821, 823, 827, 829, 839, 853, 857, 859, 863, 877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 947, 953, 967, 971, 977, 983, 991, 997, 1009, 1013, 1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063, 1069, 1087, 1091, 1093, 1097, 1103, 1109, 1117, 1123, 1129, 1151, 1153, 1163, 1171, 1181, 1187, 1193, 1201, 1213, 1217, 1223, 1229, 1231, 1237, 1249, 1259, 1277, 1279, 1283, 1289, 1291, 1297, 1301, 1303, 1307, 1319, 1321, 1327, 1361, 1367, 1373, 1381, 1399, 1409, 1423, 1427, 1429, 1433, 1439, 1447, 1451, 1453, 1459, 1471, 1481, 1483, 1487, 1489, 1493, 1499, 1511, 1523, 1531, 1543, 1549, 1553, 1559, 1567, 1571, 1579, 1583, 1597, 1601, 1607, 1609, 1613, 1619, 1621, 1627, 1637, 1657, 1663, 1667, 1669, 1693, 1697, 1699, 1709, 1721, 1723, 1733, 1741, 1747, 1753, 1759, 1777, 1783, 1787, 1789, 1801, 1811, 1823, 1831, 1847, 1861, 1867, 1871, 1873, 1877, 1879, 1889, 1901, 1907, 1913, 1931, 1933, 1949, 1951, 1973, 1979, 1987, 1993, 1997, 1999, 2003, 2011, 2017, 2027, 2029
                                            };
        #endregion
    }

    public struct DictionaryEntry<TKey, TValue> where TValue : IEquatable<TValue>
    {
        #region Properties
        private TKey _key;
        public TKey Key
        {
            get { return _key; }
            set { _key = value; }
        }

        private TValue _value;
        public TValue Value
        {
            get { return _value; }
            set { _value = value; }
        }
        #endregion

        public DictionaryEntry(TKey key, TValue val)
        {
            _key = key;
            _value = val;
        }
    }
}
